import asyncio
import attr
import logging
import importlib
from abc import ABC, abstractmethod
from datetime import datetime
from typing import Set, Union, Optional, Awaitable, Dict, Type, Tuple, List
from decimal import Decimal, getcontext
from async_property import AwaitableOnly, async_property
from privex.helpers import AttribDictable, empty, camel_to_snake, await_if_needed, dec_round, r_cache_async, empty_if
from privex.helpers.cache import async_adapter_get, async_adapter_set, AsyncCacheAdapter, AsyncMemoryCache
from privex.exchange.exceptions import PairNotFound, ProxyMissingPair
from privex.exchange.helpers import med_avg_out, avg, empty_decimal, PairListOrSet

getcontext().prec = 30

log = logging.getLogger(__name__)

ZERO, ONE, TWO = Decimal('0'), Decimal('1'), Decimal('2')
_curr_adapter = async_adapter_get()

if not isinstance(_curr_adapter, AsyncCacheAdapter):
    log.warning("WARNING: Current privex.helpers cache adapter is not an Async adapter.")
    log.warning("Current privex.helpers adapter: %s", _curr_adapter.__class__.__name__)
    log.warning("Setting privex.helpers cache adapter to 'AsyncMemoryCache'.")
    async_adapter_set(AsyncMemoryCache())

cached = async_adapter_get()
# cached = async_cached
# cached.ins_exit_close = False
# cached.ins_enter_reconnect = False


def disable_cache_context_reset():
    try:
        from privex.helpers.cache import AsyncRedisCache
        if isinstance(cached, AsyncRedisCache):
            cached.ins_exit_close = False
            cached.ins_enter_reconnect = False
        AsyncRedisCache.adapter_enter_reconnect = False
        AsyncRedisCache.adapter_exit_close = False
    except ImportError:
        pass


disable_cache_context_reset()


@attr.s
class PriceData(AttribDictable):
    """Exchange price data object returned by exchange adapters"""
    from_coin = attr.ib(type=str)
    to_coin = attr.ib(type=str)
    last = attr.ib(type=Decimal, converter=empty_decimal)
    
    bid = attr.ib(default=None, type=Decimal, converter=empty_decimal)
    ask = attr.ib(default=None, type=Decimal, converter=empty_decimal)
    
    open = attr.ib(default=None, type=Decimal, converter=empty_decimal)
    close = attr.ib(default=None, type=Decimal, converter=empty_decimal)
    
    high = attr.ib(default=None, type=Decimal, converter=empty_decimal)
    low = attr.ib(default=None, type=Decimal, converter=empty_decimal)

    volume = attr.ib(default=None, type=Decimal, converter=empty_decimal)


class ExchangeAdapter(ABC):
    """Abstract base class used by all exchange adapters for interoperability"""
    cache_timeout: int = 120
    
    _provides: Set[Tuple[str, str]] = set()
    _extra_provides: Set[Tuple[str, str]]
    validate_provides: bool
    ex_settings: dict

    def __str__(self):
        return f"[Exchange '{self.name}' (code: '{self.code}')]'"
    
    def __repr__(self):
        return self.__str__()

    @property
    @abstractmethod
    def name(self) -> str:
        return self.__class__.__name__

    @property
    @abstractmethod
    def code(self) -> str:
        return camel_to_snake(self.name)

    def __init__(self, extra_provides: PairListOrSet = None, validate_provides: bool = True, **ex_settings):
        if not empty(extra_provides, itr=True):
            for p in extra_provides:
                if self.has_pair(*p):
                    continue
                self._extra_provides.add(p)
        
        self.validate_provides = validate_provides
        self.ex_settings = dict(ex_settings)
    
    @property
    def cache(self) -> AsyncCacheAdapter:
        return cached
    
    @property
    @abstractmethod
    def provides(self) -> Union[Set[Tuple[str, str]], Awaitable[Set[Tuple[str, str]]]]:
        return set(list(self._provides) + list(self._extra_provides))

    @abstractmethod
    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        for f, t in self.provides:
            if f.upper() == from_coin.upper() and t.upper() == to_coin.upper():
                return True
        return False
    
    @abstractmethod
    async def _get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        raise NotImplemented
    
    @staticmethod
    async def _get_cache_pricedata(key: str) -> Optional[PriceData]:
        cdata = cached.get(key)
        if empty(cdata):
            return None
        
        if isinstance(cdata, PriceData):
            return cdata
    
        if isinstance(cdata, dict):
            return PriceData(**cdata)
        
        log.warning("Error reading cache key '%s' - Data was neither PriceData nor dict. Data was: %s", key, cdata)
        return None
    
    async def get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        """
        
        **Instantiate an exchange class and get a ticker asynchronously**::
        
            >>> from privex.exchange import Binance
            >>> b = Binance()
            >>> ticker = await b.get_pair('LTC', 'BTC')
            >>> ticker.last
            Decimal('0.00607100')
            >>> ticker.volume
            Decimal('89557.57000000')
        
        **Can use synchronously in non-async applications**::
        
            >>> from privex.exchange import Binance
            >>> b = Binance()
            >>> ticker = b.get_pair('LTC', 'BTC')
            >>> ticker.last
            Decimal('0.00607100')
        
        
        :param from_coin:
        :param to_coin:
        :return:
        """
        if self.validate_provides:
            if not await self.has_pair(from_coin=from_coin, to_coin=to_coin):
                raise PairNotFound(f"The coin pair '{from_coin}/{to_coin}' is not supported by {self.name}")
        
        code = self.code
        ckey = f"pvxex:{code}:{from_coin}_{to_coin}"
        with cached:
            cdata = await cached.get(ckey)
            if not empty(cdata):
                if isinstance(cdata, PriceData):
                    return cdata
            
                if isinstance(cdata, dict):
                    return PriceData(**cdata)
                log.warning("Error reading cache key '%s' - Data was neither PriceData nor dict. Data was: %s", ckey, cdata)
                log.warning(f"Will call {self.__class__.__name__}._get_pair and re-write cache key...")
            
            data = await self._get_pair(from_coin=from_coin, to_coin=to_coin)
            await cached.set(ckey, data, self.cache_timeout)
        
        return data


class ExchangeManager:
    """
    
    **Basic Usage**
    
    First create an :class:`.ExchangeManager` instance::
    
        >>> from privex.exchange import ExchangeManager
        >>> exm = ExchangeManager()
    
    Using :meth:`.get_avg` - we can get the average price for a given coin pair based on all exchanges which support that pair, as well
    as exchanges which have a proxy route (e.g. ``LTC -> BTC -> USD``)::
    
        >>> await exm.get_avg('btc', 'usd')
        Decimal('9252.50525000')
    
    You can also use :meth:`.get_pair`, which simply finds the simplest route (including proxies) via one or two exchanges to
    obtain a price::
    
        >>> await exm.get_pair('btc', 'usd')
        Decimal('9249.03000000')
        >>> await exm.get_pair('ltc', 'usd')
        Decimal('43.88000000')
    
    Converting arbitrary cryptos between each other, seamlessly::
    
        >>> await exm.get_avg('eos', 'ltc')     # LTC per 1 EOS (average via all exchanges)
        Decimal('0.0580638770094616213117276469457')
        >>> await exm.get_pair('eos', 'ltc')    # LTC per 1 EOS (direct / semi-direct)
        Decimal('0.05957304869913275517011340894')
        >>> await exm.get_avg('hive', 'eos')   # EOS per 1 HIVE
        Decimal('0.0878733984247395738811658379006')
    
    
    """
    available_adapters: List[Tuple[str, str]] = [
        ('privex.exchange.Binance', 'Binance'),
        ('privex.exchange.Bittrex', 'Bittrex'),
        ('privex.exchange.Kraken', 'Kraken'),
        ('privex.exchange.Huobi', 'Huobi'),
        ('privex.exchange.CoinGecko', 'CoinGecko'),
        ('privex.exchange.SteemEngine', 'SteemEngine'),
        ('privex.exchange.SteemEngine', 'HiveEngine'),
    ]
    
    ex_instances: Dict[str, ExchangeAdapter] = {}
    ex_code_map: Dict[str, str] = {}
    ex_name_map: Dict[str, str] = {}

    ex_pair_map: Dict[str, List[ExchangeAdapter]] = {}
    ex_pair_map_inv: Dict[ExchangeAdapter, Set[Tuple[str, str]]] = {}

    map_tether_real: bool = True
    
    tether_map = {
        'USDT': 'USD',
        'USDC': 'USD',
    }
    
    proxy_coins = ['BTC', 'USD', 'USDT', 'HIVE', 'STEEM']
    common_proxies = ['BTC', 'USD']
    _proxy_rates: Dict[str, Decimal] = {}
    proxy_rates_timestamp: Optional[datetime] = None
    proxy_rates_timeout: Union[int, float] = 300
    
    loaded = False

    def __init__(self):
        disable_cache_context_reset()
        if getcontext().prec is None or getcontext().prec < 30:
            getcontext().prec = 30

    @classmethod
    async def load_exchange(cls, package: str, name: str) -> ExchangeAdapter:
        obj_path = '.'.join([package, name])
        log.debug("Checking if adapter '%s' is already registered...", obj_path)
        
        if obj_path not in cls.ex_instances:
            log.debug("Importing module '%s'", package)
            # Try importing the package
            pkg = importlib.import_module(package)
            log.debug("Extracting object '%s' from module '%s'", name, package)
            # Then try pulling the object out of the package
            obj: Type[ExchangeAdapter] = pkg.__dict__[name]
            # Instantiate the adapter and add it to the exchange maps
            log.debug("Instantiating object '%s' and mapping it...", obj)
            inst: ExchangeAdapter = obj()
            cls.ex_instances[obj_path] = inst
            cls.ex_code_map[inst.code] = obj_path
            cls.ex_name_map[inst.name] = obj_path
            # Populate ExchangeManager.ex_pair_map
            await cls._import_pairs(inst)

        return cls.ex_instances[obj_path]

    @classmethod
    async def reload_all_pairs(cls):
        for obj_path, inst in cls.ex_instances.items():
            log.debug("Reloading pairs for adapter %s", obj_path)
            await cls._import_pairs(inst)
        log.debug("Finished reloading all provider pairs")
        return cls.ex_pair_map

    @classmethod
    async def reload_inst_pairs(cls, obj_path: str):
        inst = cls.ex_instances[obj_path]
        return await cls._import_pairs(inst)

    @classmethod
    async def _import_pairs(cls, inst: ExchangeAdapter):
        log.debug("Importing pairs for exchange adapter: %s", inst)
        # Grab the .provides set of tuples from the instance.
        _prov: Union[Set[Tuple[str, str]], Awaitable[Set[Tuple[str, str]]]] = inst.provides
        
        # Some classes may have an async_property .provides, so we need to check if it's a coroutine.
        # If it is a coroutine, then we have to await it before we can scan it.
        # noinspection PyTypeChecker
        if asyncio.iscoroutine(_prov) or asyncio.iscoroutinefunction(_prov) or isinstance(_prov, AwaitableOnly):
            log.debug(f"{inst.__class__.__name__}.provides is async. Awaiting it...")
            _prov = await _prov

        cls.ex_pair_map_inv[inst] = _prov
        # Iterate over each supported coin pair, and append the adapter instance to each pair in ex_pair_map
        for frm_coin, to_coin in _prov:
            cls._map_pair(frm_coin, to_coin, inst)

            if cls.map_tether_real:
                if frm_coin in cls.tether_map.keys():
                    log.debug(
                        "Detected frm_coin '%s' is a tethered coin. Adding extra pair map for %s/%s",
                        frm_coin, cls.tether_map[frm_coin], to_coin
                    )
                    cls._map_pair(cls.tether_map[frm_coin], to_coin, inst)
                elif to_coin in cls.tether_map.keys():
                    log.debug(
                        "Detected to_coin '%s' is a tethered coin. Adding extra pair map for %s/%s",
                        to_coin, frm_coin, cls.tether_map[to_coin]
                    )
                    cls._map_pair(frm_coin, cls.tether_map[to_coin], inst)
        
        return _prov

    @classmethod
    def list_pairs_from(cls, frm_coin: str) -> Dict[str, ExchangeAdapter]:
        return {
            pair: adapters for pair, adapters in cls.ex_pair_map.items()
            if pair.split('_')[0] == frm_coin.upper()
        }

    @classmethod
    def list_pairs_to(cls, to_coin: str) -> Dict[str, ExchangeAdapter]:
        return {
            pair: adapters for pair, adapters in cls.ex_pair_map.items()
            if pair.split('_')[1] == to_coin.upper()
        }
    
    @classmethod
    def _map_pair(cls, frm_coin: str, to_coin: str, inst: ExchangeAdapter):
        k = f"{frm_coin.upper()}_{to_coin.upper()}"
        log.debug("Mapping pair %s for exchange: %s", k, inst)
        if k in cls.ex_pair_map:
            if inst in cls.ex_pair_map[k]:
                log.debug("Exchange already exists in pair map for %s. Exchange: %s", k, inst)
                return
            log.debug("Pair %s exists in ex_pair_map. Appending exchange: %s", k, inst)
        
            cls.ex_pair_map[k].append(inst)
        else:
            log.debug("Pair %s did not exist in ex_pair_map. Creating pair with exchange: %s", k, inst)
            cls.ex_pair_map[k] = [inst]

    @classmethod
    async def load_exchanges(cls):
        coros = []
        for pkg_path, obj_name in cls.available_adapters:
            obj_path = '.'.join([pkg_path, obj_name])
            if obj_path in cls.ex_instances:
                continue
            coros.append(cls.load_exchange(pkg_path, obj_name))
            # try:
            # except:
            #     log.exception("Skipping adapter %s due to exception ...", obj_path)
        if len(coros) > 0:
            await asyncio.gather(*coros, return_exceptions=True)
        cls.loaded = True
    
    @classmethod
    def exchange_by_code(cls, code: str) -> ExchangeAdapter:
        return cls.ex_instances[cls.ex_code_map[code]]

    @classmethod
    def exchange_by_name(cls, name: str) -> ExchangeAdapter:
        return cls.ex_instances[cls.ex_name_map[name]]

    @classmethod
    def exchange_by_path(cls, obj_path: str) -> ExchangeAdapter:
        return cls.ex_instances[obj_path]

    @classmethod
    def pair_exists(cls, from_coin: str, to_coin: str, should_raise=False) -> bool:
        pair = f"{from_coin.upper()}_{to_coin.upper()}"
        if pair in cls.ex_pair_map and len(cls.ex_pair_map[pair]) > 0:
            return True
        
        if should_raise:
            raise PairNotFound(f"Pair '{pair}' does not exist / no usable exchanges in {cls.__name__}.ex_pair_map")
        
        return False
    
    @classmethod
    def _pair_exchanges(cls, from_coin: str, to_coin: str) -> List[ExchangeAdapter]:
        pair = f"{from_coin.upper()}_{to_coin.upper()}"
        if pair not in cls.ex_pair_map:
            raise PairNotFound(f"Pair '{pair}' does not exist in {cls.__name__}.ex_pair_map")
    
        if len(cls.ex_pair_map[pair]) == 0:
            raise PairNotFound(f"Pair '{pair}' has no listed exchanges in {cls.__name__}.ex_pair_map")
        
        return cls.ex_pair_map[pair]
    
    @classmethod
    async def get_proxy_avg(cls, from_coin: str, to_coin: str) -> Decimal:
        """
        Attempt to retrieve the average exchange rate for ``from_coin/to_coin`` using the proxy coin averages, which
        are retrieved and cached from :meth:`.get_proxy_rates`
        
            >>> await ExchangeManager.get_proxy_avg('BTC', 'USD')
            Decimal('9194.89249999')
        
            >>> await ExchangeManager.get_proxy_avg('HIVE', 'USD')
            Decimal('0.21594933')
            >>> await ExchangeManager.get_proxy_avg('USD', 'HIVE')
            
        :param from_coin:
        :param to_coin:
        :return:
        """
        pr = await cls.get_proxy_rates()
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        if f"{from_coin}_{to_coin}" in pr:
            return pr[f"{from_coin}_{to_coin}"]
        if f"{to_coin}_{from_coin}" in pr:
            return ONE / pr[f"{to_coin}_{from_coin}"]
        raise ProxyMissingPair(f"Could not find average proxy rate for {from_coin}/{to_coin}")
    
    @classmethod
    async def try_proxy(cls, from_coin: str, to_coin: str, proxy: str = "BTC", rate='last', adapter: ExchangeAdapter = None):
        """
        
            >>> exm = ExchangeManager()
            >>> await exm.load_exchanges()
            >>> await exm.try_proxy('HIVE', 'USD', proxy='BTC')
        
        :param from_coin:
        :param to_coin:
        :param proxy:
        :param rate:
        :param adapter:
        :return:
        """
        if not cls.loaded: await cls.load_exchanges()
        from_coin, to_coin, proxy = from_coin.upper(), to_coin.upper(), proxy.upper()
        log.info("Attempting to get '%s' price for %s / %s using proxy coin %s", rate, from_coin, to_coin, proxy)
        
        inv_to = False
        
        if adapter:
            log.debug("Custom adapter was specified. Proxying exclusively via: %s", adapter)
        
        pair_exists = cls.pair_exists if adapter is None else adapter.has_pair
        get_ticker = cls._get_ticker if adapter is None else adapter.get_pair

        if not await await_if_needed(pair_exists, from_coin, proxy):
            raise ProxyMissingPair(f"No pairs available for {from_coin} -> {proxy}")
        if not await await_if_needed(pair_exists, proxy, to_coin):
            log.debug("Did not find pair %s/%s - checking if we can do inverted %s/%s", proxy, to_coin, to_coin, proxy)
            if await await_if_needed(pair_exists, to_coin, proxy):
                log.debug("Found inverted proxy %s/%s", to_coin, proxy)
                inv_to = True
            else:
                raise ProxyMissingPair(f"No pairs available for {proxy} -> {to_coin}")
        
        # pair = f"{from_coin}_{to_coin}"
        
        # e.g. if from=HIVE, to=USD, proxy=BTC
        # get ticker for HIVE/BTC - get ticker for USD/BTC
        # e.g. HIVE/BTC = 0.00001545    BTC/USD = 6900
        # 0.00001545 * 6900 = 0.106605 USD per HIVE
        try:
            ex_from = {rate: await cls.get_proxy_avg(from_coin=from_coin, to_coin=proxy)}
        except ProxyMissingPair:
            ex_from = await get_ticker(from_coin, proxy)
        if inv_to:
            try:
                ex_to = {rate: await cls.get_proxy_avg(from_coin=to_coin, to_coin=proxy)}
            except ProxyMissingPair:
                ex_to = await get_ticker(to_coin, proxy)
            log.debug("Original rates from reverse proxy %s/%s - rates: %s", to_coin, proxy, ex_to)
            ex_to = ExchangeManager._invert_data(ex_to)
            log.debug("Inverted rates - converted pair into %s/%s : %s", proxy, to_coin, ex_to)
        else:
            try:
                ex_to = {rate: await cls.get_proxy_avg(from_coin=proxy, to_coin=to_coin)}
            except ProxyMissingPair:
                ex_to = await get_ticker(proxy, to_coin)

        r_from, r_to = Decimal(ex_from[rate]), Decimal(ex_to[rate])
        
        log.debug(f"{from_coin}/{proxy} price: {r_from:.8f} {proxy} per 1 {from_coin}")
        log.debug(f"{proxy}/{to_coin} price: {r_to:.8f} {to_coin} per 1 {proxy}")
        log.debug(f"{r_from} * {r_to} = %f", r_from * r_to)
        return r_from * r_to
    
    @staticmethod
    def _invert_data(data: PriceData):
        inv_keys = ['last', 'bid', 'ask', 'open', 'close', 'high', 'low', 'volume']
        new_pd = PriceData(**dict(data))
        
        # Invert each ticker price, e.g. for BTC/USD into USD/BTC
        for k in inv_keys:
            if empty(new_pd[k], zero=True):
                continue
            setattr(new_pd, k, Decimal('1') / new_pd[k])
        
        # Atomic swap from_coin and to_coin with eachother
        new_pd.from_coin, new_pd.to_coin = new_pd.to_coin, new_pd.from_coin
        
        return new_pd
    
    @async_property
    async def proxy_rates(self) -> Dict[str, Decimal]:
        """This is just an AsyncIO property for :meth:`.get_proxy_rates`"""
        return await self.get_proxy_rates()
    
    @classmethod
    @r_cache_async(lambda cls: f"pvxex:exm:proxy_rates:{','.join(cls.proxy_coins)}", cache_time=proxy_rates_timeout)
    async def get_proxy_rates(cls) -> Dict[str, Decimal]:
        """
        Retrieves and caches proxy rates from :meth:`.prep_proxy_rates`
        """
        pr = await cls.prep_proxy_rates()
        return pr
    
    @classmethod
    async def prep_proxy_rates(cls) -> Dict[str, Decimal]:
        """
        Internal usage. Initialises / updates :attr:`._proxy_rates` with string coin pairs mapped to their exchange rate.
        
        
        Sets :attr:`._proxy_rates` and returns a dictionary of proxy coin pairs mapped to their ex rate::
        
            {'BTC_USD': Decimal('9183.20224999'), 'BTC_USDT': Decimal('9190.78333333'), 'BTC_HIVE': Decimal('42500.00000000'),
             'BTC_STEEM': Decimal('44900.00000000'), 'USD_USDT': Decimal('1.00085000'), 'USDT_BTC': Decimal('0.00010869'),
             'USDT_USD': Decimal('0.99920000'), 'HIVE_BTC': Decimal('0.00002362'), 'HIVE_USD': Decimal('0.21596266'),
             'HIVE_USDT': Decimal('0.21577000'), 'HIVE_STEEM': Decimal('1.06000100'), 'STEEM_BTC': Decimal('0.00002231'),
             'STEEM_USD': Decimal('0.20394399'), 'STEEM_HIVE': Decimal('0.92999999')
             }
        
        """
        if not cls.loaded: await cls.load_exchanges()
        
        # ex_rates: Dict[str, Dict[str, Decimal]] = {}
        ex_rates_avg: Dict[str, Decimal] = {}
        
        rate_coros = []
        
        async def x_rates(f, t):
            rates = await cls.get_all_rates(from_coin=f, to_coin=t)
            return f"{f}_{t}", rates
        
        for frm in cls.proxy_coins:
            for to in cls.proxy_coins:
                if frm == to: continue
                rate_coros.append(x_rates(frm, to))
                # pair = f"{frm}_{to}"
                # for _, adp in cls.ex_instances.items():
                #     try:
                #         if not await adp.has_pair(frm, to): continue
                #         if pair not in ex_rates:
                #             ex_rates[pair] = {}
                #         log.debug(f"Getting proxy rate for {pair} from exchange '{adp.code}'")
                #         pd = await adp.get_pair(from_coin=frm, to_coin=to)
                #         if empty(pd.last, zero=True):
                #             raise ValueError(f"Price data 'last' from exchage '{adp.code}' was empty / zero!")
                #         ex_rates[pair][adp.code] = pd.last
                #     except Exception as e:
                #         log.warning(f"Failed to get {pair} from exchange '{adp.code}'. Reason: %s %s", type(e), str(e))
        
        rate_res = await asyncio.gather(*rate_coros, return_exceptions=True)
        # for p, rates in ex_rates.items():
        for res in rate_res:
            if not isinstance(res, tuple): continue
            
            if len(res[1].values()) < 1:
                # log.debug("Pair %s has no values. Skipping.", p)
                continue
            ex_rates_avg[res[0]] = med_avg_out(*res[1].values())
            log.debug("Average rate for %s is %f - based on: %s", res[0], ex_rates_avg[res[0]], res[1])
        
        log.debug("Got proxy averages: %s", ex_rates_avg)
        cls._proxy_rates = ex_rates_avg
        return cls._proxy_rates
    
    @classmethod
    async def _get_ticker(cls, from_coin: str, to_coin: str, ex_index: int = 0) -> PriceData:
        if not cls.loaded: await cls.load_exchanges()
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        
        try:
            cls.pair_exists(from_coin, to_coin, should_raise=True)
        except PairNotFound:
            log.info("Pair %s/%s was not found. Trying inverse query using %s/%s", from_coin, to_coin, to_coin, from_coin)
            cls.pair_exists(to_coin, from_coin, should_raise=True)
            
            adp = cls.ex_pair_map[f"{to_coin}_{from_coin}"][ex_index]
            return ExchangeManager._invert_data(await adp.get_pair(to_coin, from_coin))
        
        adp = cls.ex_pair_map[f"{from_coin}_{to_coin}"][ex_index]
        log.info("Found pair %s/%s - querying exchange: %s", from_coin, to_coin, adp)
        return await adp.get_pair(from_coin, to_coin)

    @classmethod
    async def get_tickers(cls, from_coin: str, to_coin: str, rate='last') -> Dict[str, Decimal]:
        if not cls.loaded: await cls.load_exchanges()
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        data = {}
        
        for _, adp in cls.ex_instances.items(): # type: ExchangeAdapter
            if await adp.has_pair(from_coin, to_coin):
                try:
                    d = await adp.get_pair(from_coin, to_coin)
                    data[adp.code] = d[rate]
                except Exception as e:
                    log.warning(
                        "Error obtaining %s/%s exchange rate from %s - reason: %s %s",
                        from_coin, to_coin, adp.name, type(e), str(e)
                    )
            elif await adp.has_pair(to_coin, from_coin):
                log.info("Pair %s/%s was not found. Trying inverse query using %s/%s", from_coin, to_coin, to_coin, from_coin)
                try:
                    data[adp.code] = ExchangeManager._invert_data(await adp.get_pair(to_coin, from_coin))[rate]
                except Exception as e:
                    log.warning(
                        "Error obtaining %s/%s exchange rate from %s - reason: %s %s",
                        from_coin, to_coin, adp.name, type(e), str(e)
                    )
            else:
                try:
                    proxy = await cls._find_proxy(from_coin, to_coin, adapter=adp)
                    data[adp.code] = await cls.try_proxy(from_coin, to_coin, proxy, adapter=adp)
                except (ProxyMissingPair, PairNotFound):
                    log.warning(
                        "Error obtaining %s/%s exchange rate from %s - cannot find a proxy pair.",
                        from_coin, to_coin, adp.name
                    )
            
        return data

    @classmethod
    async def _find_proxy(cls, from_coin: str, to_coin: str, adapter: ExchangeAdapter = None) -> str:
        """
        
            >>> await cls._find_proxy('HIVE', 'USD')
            'BTC'
        
        :param from_coin:
        :param to_coin:
        :return:
        """
        if not cls.loaded: await cls.load_exchanges()
        from_coin, to_coin = from_coin.upper(), to_coin.upper()

        for c in cls.proxy_coins:
            if adapter:
                if not await adapter.has_pair(from_coin, c): continue
                if await adapter.has_pair(c, to_coin): return c
                if await adapter.has_pair(to_coin, c): return c
                continue
            
            if f"{from_coin}_{c}" not in cls.ex_pair_map:
                log.debug("Cannot proxy via %s - pair %s not found", c, f"{from_coin}_{c}")
                continue
            
            if f"{c}_{to_coin}" in cls.ex_pair_map:
                log.debug("Found forward proxy via %s - pair %s found", c, f"{c}_{to_coin}")
                return c
            
            if f"{to_coin}_{c}" in cls.ex_pair_map:
                log.debug("Found reverse proxy via %s - pair %s found", c, f"{to_coin}_{c}")
                return c
        raise ProxyMissingPair(f"Could not find a viable proxy route for {from_coin} -> {to_coin}")
    
    @classmethod
    async def get_all_rates(cls, from_coin: str, to_coin: str, rate='last', invert=False):
        if not cls.loaded: await cls.load_exchanges()
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        # frm, to = from_coin, to_coin
        pair = f"{from_coin}_{to_coin}"
        ex_rates = {}
        
        async def _get_rate(frm, to, adp, rate_key):
            try:
                if not await adp.has_pair(frm, to): return None
                # if pair not in ex_rates: ex_rates[pair] = {}
                log.debug(f"Getting rate for {pair} from exchange '{adp.code}'")
                pd = await adp.get_pair(from_coin=frm, to_coin=to)
                if empty(pd.last, zero=True):
                    raise ValueError(f"Price data 'last' from exchange '{adp.code}' was empty / zero!")
                # ex_rates[adp.code] = pd[rate]
                return adp.code, pd[rate_key]
            except Exception as e:
                log.warning(f"Failed to get {pair} from exchange '{adp.code}'. Reason: %s %s", type(e), str(e))
                return None
        
        rate_tasks = [_get_rate(from_coin, to_coin, adapter, rate) for adapter in cls.ex_instances.values()]
        rate_res = await asyncio.gather(*rate_tasks, return_exceptions=True)
        
        for r in rate_res:
            if not isinstance(r, tuple): continue
            ex_rates[r[0]] = ONE / r[1] if invert else r[1]
        # for _, adapter in cls.ex_instances.items():
        #     try:
        #         if not await adp.has_pair(frm, to): continue
        #         # if pair not in ex_rates: ex_rates[pair] = {}
        #         pd = await adp.get_pair(from_coin=frm, to_coin=to)
        #         if empty(pd.last, zero=True):
        #             raise ValueError(f"Price data 'last' from exchage '{adp.code}' was empty / zero!")
        #         ex_rates[adp.code] = pd[rate]
        #         log.debug(f"Getting rate for {pair} from exchange '{adp.code}'")
        #     except Exception as e:
        #         log.warning(f"Failed to get {pair} from exchange '{adp.code}'. Reason: %s %s", type(e), str(e))
        log.debug("All rates (inverted: %s) for %s/%s: %s", invert, from_coin, to_coin, ex_rates)
        return ex_rates
    
    @classmethod
    async def get_direct_rate_avg(cls, from_coin: str, to_coin: str, rate='last', invert=False) -> Optional[Decimal]:
        if rate == 'last':
            try:
                r = await cls.get_proxy_avg(from_coin, to_coin)
                return ONE / r if invert else r
            except ProxyMissingPair:
                pass
        # proxy_rates = await cls.get_proxy_rates()
        # pair, inv_pair = f"{from_coin}_{to_coin}", f"{to_coin}_{from_coin}"
        # if pair in proxy_rates:
        #     return ONE / proxy_rates[pair] if invert else proxy_rates[pair]
        # if inv_pair in proxy_rates:
        #     return proxy_rates[inv_pair] if invert else ONE / proxy_rates[inv_pair]
        
        rates = await cls.get_all_rates(from_coin, to_coin, rate=rate, invert=invert)
        if len(rates) < 1:
            return None
        return med_avg_out(*rates.values())
    
    @classmethod
    async def _get_avg_proxy(cls, from_coin: str, to_coin: str, proxy: str = 'BTC', fail=False, rate='last') -> Optional[Decimal]:
        if not cls.loaded: await cls.load_exchanges()
        from_coin, to_coin, proxy = from_coin.upper(), to_coin.upper(), proxy.upper()
        
        # To make the below proxy routes easier to understand, they're commented with examples
        # based on: from_coin='HIVE', to_coin='USD', proxy='BTC' (HIVE/USD via BTC proxy)
        
        # HIVE/BTC -> BTC/USD
        if cls.pair_exists(from_coin, proxy) and cls.pair_exists(proxy, to_coin):
            avg_frm = await cls.get_direct_rate_avg(from_coin, proxy, rate=rate)  # HIVE -> BTC
            avg_to = await cls.get_direct_rate_avg(proxy, to_coin, rate=rate)  # BTC -> USD
        # HIVE/BTC -> USD/BTC
        elif cls.pair_exists(from_coin, proxy) and cls.pair_exists(to_coin, proxy):
            avg_frm = await cls.get_direct_rate_avg(from_coin, proxy, rate=rate)
            avg_to = ONE / (await cls.get_direct_rate_avg(to_coin, proxy, rate=rate))
        # BTC/HIVE -> BTC/USD
        elif cls.pair_exists(proxy, from_coin) and cls.pair_exists(proxy, to_coin):
            avg_frm = ONE / (await cls.get_direct_rate_avg(proxy, from_coin, rate=rate))
            avg_to = await cls.get_direct_rate_avg(proxy, to_coin, rate=rate)
        # BTC/HIVE -> USD/BTC
        elif cls.pair_exists(proxy, from_coin) and cls.pair_exists(to_coin, proxy):
            avg_frm = ONE / (await cls.get_direct_rate_avg(proxy, from_coin, rate=rate))
            avg_to = ONE / (await cls.get_direct_rate_avg(to_coin, proxy, rate=rate))
            
        else:
            log.info("Failed to find a proxy for %s / %s via %s", from_coin, to_coin, proxy)
            if fail:
                raise ProxyMissingPair(f"Could not find any route to get {from_coin} -> {to_coin} via {proxy}")
            return None
        
        return avg_frm * avg_to

    @classmethod
    async def _get_avg_proxies(cls, from_coin: str, to_coin: str, *proxies, fail=False, rate='last') -> Optional[Decimal]:
        proxies = [v.upper() for v in empty_if(proxies, cls.common_proxies, itr=True)]
        for c in proxies:
            avg_rate = await cls._get_avg_proxy(from_coin, to_coin, proxy=c, fail=False, rate=rate)
            if not empty(avg_rate, True, True):
                return avg_rate
        if fail:
            raise ProxyMissingPair(f"Could not find any route to get {from_coin} -> {to_coin} via proxies: {proxies}")
        return None
    
    async def get_avg(self, from_coin: str, to_coin: str, use_proxy=True, rate='last', dp: int = 8) -> Decimal:
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        if not self.loaded:
            await self.load_exchanges()
        
        # async def get_inverted(frm, to, xrate):
        #     log.debug()
        #     x = await self.get_direct_rate_avg(to, frm, rate=xrate)
        #     if empty(x): return None
        #     return ONE / x
        
        coros = []
        
        if use_proxy:
            coros.append(self._get_avg_proxies(from_coin, to_coin, fail=True, rate=rate))
        
        pair = f"{from_coin}_{to_coin}"
        if pair in self.ex_pair_map:
            coros.append(self.get_direct_rate_avg(from_coin, to_coin, rate=rate))
        elif f"{to_coin}_{from_coin}" in self.ex_pair_map:
            coros.append(self.get_direct_rate_avg(to_coin, from_coin, rate=rate, invert=True))
        
        # if pair not in self.ex_pair_map:
        #     if not use_proxy:
        #         raise PairNotFound(f"Pair {pair} was not found in ex_pair_map, and no proxying requested.")
        #     log.debug("%s not found in pair map - only using proxies", pair)
        #     avg_rate = await self._get_avg_proxies(from_coin, to_coin, fail=True, rate=rate)
        #     return avg_rate
        # log.debug("%s WAS found in pair map - using both proxies and direct rates")
        #
        # _rates = await asyncio.gather(self._get_avg_proxies(from_coin, to_coin, fail=False, rate=rate),
        #                               self.get_direct_rate_avg(from_coin, to_coin, rate=rate))
        _rates = await asyncio.gather(*coros, return_exceptions=True)
        return avg(*_rates, dp=dp)
        # proxy_rate = await self._get_avg_proxies(from_coin, to_coin, fail=False, rate=rate)
        # direct_rate = await self.get_direct_rate_avg(from_coin, to_coin, rate=rate)
        # if empty(proxy_rate, True):
        #     return direct_rate
        # if empty(direct_rate, True):
        #     return proxy_rate
        # combined_avg = (proxy_rate + direct_rate) / TWO
        # return dec_round(combined_avg, dp=8)
        
    async def get_pair(self, from_coin: str, to_coin: str, rate='last', use_proxy: bool = True, dp: int = 8):
        if not self.loaded:
            await self.load_exchanges()
        
        pair = f"{from_coin.upper()}_{to_coin.upper()}"
        
        if pair not in self.ex_pair_map:
            if not use_proxy:
                raise PairNotFound(f"Pair {pair} was not found in ex_pair_map, and no proxying requested.")
            try:
                log.debug("Pair %s not found - trying to find a proxy route...", pair)
                proxy = await self._find_proxy(from_coin, to_coin)
                log.debug("Found proxy to %s via %s - trying proxy...", to_coin, proxy)
                rate_proxy = await self.try_proxy(from_coin, to_coin, proxy, rate=rate)
                log.debug("Got proxy rate: %f %s per %s", rate_proxy, to_coin, from_coin)
                return rate_proxy
            except ProxyMissingPair:
                raise PairNotFound(f"Pair {pair} not found, nor a viable proxy route.")
        
        adp = self.ex_pair_map[pair][0]
        log.debug("Pair %s found - getting exchange rate directly", pair)

        data = await adp.get_pair(from_coin, to_coin)
        return dec_round(data[rate], dp=dp)


class AsyncProvidesAdapter(ExchangeAdapter, ABC):
    code: str
    cache: AsyncCacheAdapter
    _gen_provides: callable
    
    PROVIDES_CACHE_TIME: Union[int, float] = 300
    """Amount of time (in seconds) to cache :attr:`.provides` for"""
    
    @async_property
    async def provides(self) -> Set[Tuple[str, str]]:
        """
        Kraken provides ALL of their tickers through one GET query, so we can generate ``provides``
        simply by querying their API via :meth:`.load_pairs` (using :meth:`._gen_provides`)

        We cache the provides Set both class-locally in :attr:`._provides`, as well as via the Privex Helpers
        Cache system - :mod:`privex.helpers.cache`

        """
        if empty(self._provides, itr=True):
            log.debug("Getting _provides from cache for %s", self.code)
            _prov = await self.cache.get(f"pvxex:{self.code}:provides")
            if not empty(_prov):
                # log.debug("Got _provides from cache")
                self._provides = _prov
            else:
                # log.debug("Calling _gen_provides for %s", self.code)
                self._provides = await self._gen_provides()
                # log.debug("Adding _provides to cache for %s : %s", self.code, self._provides)
                await self.cache.set(f"pvxex:{self.code}:provides", self._provides, timeout=self.PROVIDES_CACHE_TIME)
        # log.debug("Returning _provides")
        return self._provides


__all__ = [
    'ExchangeAdapter', 'PriceData',
    'ExchangeManager', 'AsyncProvidesAdapter'
]



from json import JSONDecodeError
from typing import Set, Tuple, Optional, AsyncGenerator, Dict, List
from async_property import async_property
from httpx import HTTPError
from privex.helpers import cached, empty
from privex.exchange.base import ExchangeAdapter, PriceData
from privex.exchange.exceptions import PairNotFound, ExchangeDown
import httpx

import logging

log = logging.getLogger(__name__)


# @awaitable_class
class ExampleExchange(ExchangeAdapter):
    TICKER_API = 'https://example.com/api/v1/ticker'
    
    name = "Example Exchange"
    code = "example"
    
    _provides = set()
    _extra_provides = set()
    
    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        prov = await self.provides
        pairs = []
        for f, t in prov:
            pairs += [f"{f.upper()}_{t.upper()}"]
        
        from_coin, to_coin, = from_coin.upper(), to_coin.upper()
        
        if f"{from_coin}_{to_coin}" in pairs:
            return True
        
        if from_coin == 'USD' and (f"USDT_{to_coin}" in pairs or f"USDC_{to_coin}" in pairs):
            return True
        if to_coin == 'USD' and (f"{from_coin}_USDT" in pairs or f"{from_coin}_USDC" in pairs):
            return True
        
        return False
    
    @async_property
    async def tickers(self) -> Dict[str, PriceData]:
        return await self.get_tickers()
    
    async def get_tickers(self) -> Dict[str, PriceData]:
        # First try and get the ticker data from cache
        ckey = f"pvxex:{self.code}:all_tickers"
        cdata = await cached.get(ckey)
        
        if not empty(cdata):
            return cdata
        
        # If we don't have it in the privex-helpers cache, query the API, cache the data and return it.
        data = {}
        async for pd in self._get_tickers():
            # Makes a dictionary map of pair > PriceData
            # e.g. data['BTC_USD'] = PriceData(from_coin='BTC', to_coin='USD', last=Decimal('9001.123'))
            data[f"{pd.from_coin}_{pd.to_coin}"] = pd
        
        await cached.set(ckey, data)
        
        return data
    
    async def _get_tickers(self) -> AsyncGenerator[PriceData, None]:
        """

        :return AsyncGenerator[PriceData,None] ticker: An async generator of :class:`.PriceData` tickers
        """
        # Query the Binance API and obtain the ticker data as a list of dictionaries
        async with httpx.AsyncClient() as client:
            data = await client.get(self.TICKER_API, timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
            res: List[dict] = data.json()
        # Loop over each ticker dictionary and return a PriceData object
        for d in res:  # type: dict
            yield  # PriceData()
            # yield PriceData(
            #     from_coin=sym.split(b)[0], to_coin=b, last=d.get('lastPrice'), bid=d.get('bidPrice'),
            #     ask=d.get('askPrice'), open=d.get('openPrice'), close=d.get('prevClosePrice'),
            #     high=d.get('highPrice'), low=d.get('lowPrice'), volume=d.get('volume')
            # )
    
    # noinspection PyTypeChecker
    async def _gen_provides(self, *args, **kwargs) -> Set[Tuple[str, str]]:
        return
    
    @async_property
    async def provides(self) -> Set[Tuple[str, str]]:
        """
        We cache the provides Set both class-locally in :attr:`._provides`, as well as via the Privex Helpers
        Cache system - :mod:`privex.helpers.cache`

        """
        if empty(self._provides, itr=True):
            _prov = await cached.get(f"pvxex:{self.code}:provides")
            if not empty(_prov):
                self._provides = _prov
            else:
                self._provides = await self._gen_provides()
                await cached.set(f"pvxex:{self.code}:provides", self._provides)
        return self._provides
    
    async def _get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        pass





from json import JSONDecodeError
from typing import Set, Tuple, Optional, AsyncGenerator, Dict, List

from async_property import async_property
from httpx import HTTPError
from privex.helpers import empty, awaitable_class
# from privex.helpers.cache import adapter_get

from privex.exchange.base import AsyncProvidesAdapter, PriceData
from privex.exchange.exceptions import PairNotFound, ExchangeDown
import httpx
import logging

log = logging.getLogger(__name__)
# cached = adapter_get(AsyncMemoryCache)


# @awaitable_class
class Binance(AsyncProvidesAdapter):
    TICKER_API = 'https://api.binance.com/api/v1/ticker/24hr'
    
    name = "Binance"
    code = "binance"
    
    known_bases = [
        "BTC", "USDT", "USDC",
        "BUSD", "TUSD", "USD",
        "ETH", "TRX", "XRP",
        "PAX", "BKRW", "EUR",
        "NGN", "RUB", "TRY",
        "ZAR"
    ]
    
    _provides = set()
    _extra_provides = set()

    def _find_base(self, pair: str) -> Optional[str]:
        pair = pair.upper()
        for b in self.known_bases:
            if pair.endswith(b):
                return b
        return None

    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        prov = await self.provides
        pairs = []
        for f, t in prov:
            pairs += [f"{f.upper()}_{t.upper()}"]

        from_coin, to_coin, = from_coin.upper(), to_coin.upper()
        
        if f"{from_coin}_{to_coin}" in pairs:
            return True
        
        if from_coin == 'USD' and (f"USDT_{to_coin}" in pairs or f"USDC_{to_coin}" in pairs):
            return True
        if to_coin == 'USD' and (f"{from_coin}_USDT" in pairs or f"{from_coin}_USDC" in pairs):
            return True

        return False

    @async_property
    async def tickers(self) -> Dict[str, PriceData]:
        return await self.get_tickers()

    async def get_tickers(self) -> Dict[str, PriceData]:
        # First try and get the ticker data from cache
        ckey = f"pvxex:{self.code}:all_tickers"
        cdata = await self.cache.get(ckey)
        
        if not empty(cdata):
            return cdata
        
        # If we don't have it in the privex-helpers cache, query the Binance API, cache the data and return it.
        data = {}
        async for pd in self._get_tickers():
            # Makes a dictionary map of pair > PriceData
            # e.g. data['BTC_USD'] = PriceData(from_coin='BTC', to_coin='USD', last=Decimal('9001.123'))
            data[f"{pd.from_coin}_{pd.to_coin}"] = pd
            # if pd.from_coin == 'USDT':
            #     data[f"USD_{pd.to_coin}"] = pd
            # elif pd.to_coin == 'USDT':
            #     data[f"{pd.from_coin}_USD"] = pd

        await self.cache.set(ckey, data)
        
        return data

    async def _get_tickers(self) -> AsyncGenerator[PriceData, None]:
        """
        Used internally by :meth:`.get_tickers`
        
        Queries the Binance 24hr ticker API :attr:`.TICKER_API` asynchronously, and returns
        :class:`.PriceData` objects as an async generator (``async for x in self._get_tickers()``)
        
        :return AsyncGenerator[PriceData,None] ticker: An async generator of :class:`.PriceData` tickers
        """
        # Query the Binance API and obtain the ticker data as a list of dictionaries
        async with httpx.AsyncClient() as client:
            data = await client.get(self.TICKER_API, timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
            res: List[dict] = data.json()
        # Loop over each ticker dictionary and return a PriceData object
        for d in res:   # type: dict
            sym: str = d['symbol']
            # Binance pairs are formatted like 'BTCUSD', so we need to scan the pair to figure out what
            # the 'to_coin' symbol actually is, then we can extract the from_coin.
            b = self._find_base(sym)
            if b is None:
                log.debug("Skipping symbol '%s' as could not identify base symbol of pair...", sym)
                continue
            
            yield PriceData(
                from_coin=sym.split(b)[0], to_coin=b, last=d.get('lastPrice'), bid=d.get('bidPrice'),
                ask=d.get('askPrice'), open=d.get('openPrice'), close=d.get('prevClosePrice'),
                high=d.get('highPrice'), low=d.get('lowPrice'), volume=d.get('volume')
            )

    # noinspection PyTypeChecker
    async def _gen_provides(self, *args, **kwargs) -> Set[Tuple[str, str]]:
        t = await self.get_tickers()
        _provides: Set[Tuple[str, str]] = set()
        for k, _ in t.items():
            _provides.add(tuple(k.split('_')))
        return _provides

    # @async_property
    # async def provides(self) -> Set[Tuple[str, str]]:
    #     """
    #     Binance provides ALL of their tickers through one GET query, so we can generate ``provides``
    #     simply by querying their API via :meth:`._get_tickers` (using :meth:`._gen_provides`)
    #
    #     We cache the provides Set both class-locally in :attr:`._provides`, as well as via the Privex Helpers
    #     Cache system - :mod:`privex.helpers.cache`
    #
    #     """
    #     if empty(self._provides, itr=True):
    #         _prov = await self.cache.get(f"pvxex:{self.code}:provides")
    #         if not empty(_prov):
    #             self._provides = _prov
    #         else:
    #             self._provides = await self._gen_provides()
    #             await self.cache.set(f"pvxex:{self.code}:provides", self._provides)
    #     return self._provides

    async def _get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        tickers = await self.tickers
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        key = f"{from_coin}_{to_coin}"
        if key not in tickers:
            if from_coin == 'USD': key = f"USDT_{to_coin}"
            if to_coin == 'USD': key = f"{from_coin}_USDT"
            if key in tickers: return tickers[key]
            
            if from_coin == 'USD': key = f"USDC_{to_coin}"
            if to_coin == 'USD': key = f"{from_coin}_USDC"
            if key in tickers: return tickers[key]
            
            raise PairNotFound(f"The coin pair '{from_coin}/{to_coin}' is not supported by {self.name}")
        
        return tickers[key]
        




from json import JSONDecodeError
from typing import Set, Tuple, Optional, AsyncGenerator, Dict, List, Union

from async_property import async_property
from httpx import HTTPError
from privex.helpers import empty, awaitable_class, r_cache_async

from privex.exchange.base import PriceData, AsyncProvidesAdapter
from privex.exchange.exceptions import PairNotFound, ExchangeDown
import httpx

import logging

log = logging.getLogger(__name__)

# cached = adapter_get()


# @awaitable_class
class Kraken(AsyncProvidesAdapter):
    TICKER_API = 'https://api.kraken.com/0/public'
    
    name = "Kraken"
    code = "kraken"

    symbol_map_expected = {
        'DOGE': ['XDG', 'XXDG'],
        'BTC': ['XXBT', 'XBT'],
        'XBT': ['XXBT', 'XBT'],
        'LTC': ['XLTC', 'LTC'],
        'ETH': ['XETH', 'ETH'],
        'ETC': ['XETC', 'ETC'],
        'XRP': ['XXRP', 'XRP'],
        'XMR': ['XXMR', 'XMR'],
        'USD': ['ZUSD', 'USD', 'USDT', 'USDC'],
        'EUR': ['ZEUR', 'EUR'],
        'GBP': ['ZGBP', 'GBP'],
        'CAD': ['ZCAD', 'CAD'],
        'JPY': ['ZJPY', 'JPY'],
    }
    """
    Kraken's asset pairs are very inconsistent, so this dictionary maps "sane" symbols, to the various symbols
    kraken's asset pairs use, allowing :func:`.get_ticker` to "guess" the symbol combination by trying each
    symbol until it figures out the correct symbol combination for a pair.
    """
    symbol_map = {
        'XXDG': 'DOGE',
        'XDG': 'DOGE',
        'XXBT': 'BTC',
        'XBT': 'BTC',
        'XLTC': 'LTC',
        'ZUSD': 'USD',
        'ZEUR': 'EUR',
        'ZGBP': 'GBP',
        'ZJPY': 'JPY',
    }

    @property
    def symbol_map_inverted(self) -> List[Tuple[str, str]]:
        return [
            (y, x) for x, y in self.symbol_map.items()
        ]

    known_bases = list(symbol_map.keys()) + [
        "BTC", "ETH", "USDT", "USDC",
        "USD", "GBP", "EUR", "JPY",
        "CAD", "CHF", "DAI"
    ]
    
    known_pairs = {
        "BTC_USD": "XXBTZUSD", "LTC_USD": "XLTCZUSD", "ETH_USD": "XETHZUSD",
        "BTC_EUR": "XXBTZEUR", "LTC_EUR": "XLTCZEUR", "ETH_EUR": "XETHZEUR",
        "BTC_GBP": "XXBTZGBP", "LTC_GBP": "XLTCZGBP", "ETH_GBP": "XETHZGBP",
        "EOS_USD": "EOSUSD", "EOS_BTC": "EOSXBT",
        "LTC_BTC": "XLTCXXBT", "ETH_BTC": "XETHXXBT",
        "USD_EUR": "USDTEUR", "USD_GBP": "USDTGBP", "USD_CAD": "USDTCAD",
    }
    """
    Kraken has extremely inconsistent pairs, so we map some of the most common pairs in "sane" format,
    to the format Kraken expects.
    
    If the class doesn't have a registered "known pair", it falls back to brute forcing the pair
    via :attr:`.symbol_map_expected`
    """
    
    _provides = set()
    _extra_provides = set()

    async def _query(self, endpoint='') -> Union[list, dict]:
        url = f"{self.TICKER_API}/{endpoint}"
        async with httpx.AsyncClient() as client:
            data = await client.get(url, timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
        j = data.json()
        if not empty(j.get('error'), zero=True, itr=True):
            raise ExchangeDown(f"Error querying {self.name} URL '{url}' - Error is: {j['error']}")
        return j
    
    def _find_base(self, pair: str) -> Optional[str]:
        pair = pair.upper()
        for b in self.known_bases:
            if pair.endswith(b):
                return b
        return None
    
    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        prov = await self.provides
        pairs = []
        for f, t in prov:
            pairs += [f"{f.upper()}_{t.upper()}"]
        
        from_coin, to_coin, = from_coin.upper(), to_coin.upper()
        # if from_coin == 'BTC': from_coin = 'XBT'
        # if to_coin == 'BTC': to_coin = 'XBT'
        if f"{from_coin}_{to_coin}" in pairs:
            return True

        if to_coin == 'BTC' and f"{from_coin}_XBT" in pairs:
            return True
        if from_coin == 'USD' and (f"USDT_{to_coin}" in pairs or f"USDC_{to_coin}" in pairs):
            return True
        if to_coin == 'USD' and (f"{from_coin}_USDT" in pairs or f"{from_coin}_USDC" in pairs):
            return True
        
        return False

    @r_cache_async(f"pvxex:kraken:all_pairs", 300)
    async def load_pairs(self) -> List[str]:
        # If we don't have it in the privex-helpers cache, query the Bittrex API, cache the data and return it.
        data = []
        log.debug("[load_pairs] Loading %s pairs via _load_pairs", self.code)
        pairs = await self._load_pairs()
        for pair in pairs:
            # Kraken pairs are formatted like 'BTCUSD', so we need to scan the pair to figure out what
            # the 'to_coin' symbol actually is, then we can extract the from_coin.
            b = self._find_base(pair)
            if b is None:
                log.debug("Skipping symbol '%s' as could not identify base symbol of pair...", pair)
                continue
            from_coin, to_coin = pair.split(b)[0], b
            if b in self.symbol_map:
                to_coin = self.symbol_map[b]
            if from_coin in self.symbol_map:
                from_coin = self.symbol_map[from_coin]
            log.debug("Adding %s pair '%s_%s'", self.code, from_coin.upper(), to_coin.upper())
            data += [f"{from_coin.upper()}_{to_coin.upper()}"]
    
        # await cached.set(ckey, data)
        log.debug("Returning pair data: %s", data)
        return data

    # async def _load_pairs(self) -> AsyncGenerator[str, None]:
    async def _load_pairs(self) -> List[str]:
        """
        Used internally by :meth:`.get_tickers`

        Queries the Bittrex market API :attr:`.MARKET_API` asynchronously, and returns
        ``Tuple[str, str]`` objects as an async generator::

            >>> for frm_coin, to_coin in self._load_pairs():
            ...     print(frm_coin, to_coin)

        :return AsyncGenerator[PriceData,None] ticker: An async generator of ticker pairs
        """
        res = await self._query('AssetPairs?info=fees')
        # Loop over each ticker dictionary and return the pair
        pairs = []
        for k, v in res['result'].items():
            # yield k
            pairs.append(k)
        return pairs
    
    async def _get_ticker(self, pair: str):
        res = await self._query(f'Ticker?pair={pair}')
        res: dict = res['result']
        res = res[list(res.keys())[0]]
        
        return PriceData(
            from_coin=pair, to_coin=pair,
            ask=res['a'][0], bid=res['b'][0], last=res['c'][0],
            open=res['o'], low=res['l'][0], high=res['h'][0],
            volume=res['v'][0]
        )

    async def get_ticker(self, from_coin: str, to_coin: str) -> PriceData:
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        if f"{from_coin}_{to_coin}" in self.known_pairs:
            
            xpair = self.known_pairs[f"{from_coin}_{to_coin}"]
            log.debug(
                "Using known Kraken pair for %s/%s: '%s'",
                from_coin, to_coin, xpair
            )
            return await self._get_ticker(xpair)
        
        # If we don't have "expected" kraken symbols mapped for the given coins, fallback to the user specified symbols
        from_coins = [from_coin]
        to_coins = [to_coin]
        
        # If we know kraken expects certain symbols for from_coin/to_coin, use those instead of the user's symbol.
        if from_coin in self.symbol_map_expected:
            from_coins = self.symbol_map_expected[from_coin]
        if to_coin in self.symbol_map_expected:
            to_coins = self.symbol_map_expected[to_coin]
        
        # "brute force" the correct Kraken pair, because Kraken's pairs are very inconsistent :)
        for fc in from_coins:
            for tc in to_coins:
                try:
                    log.debug("Trying guessed kraken pair %s/%s", fc, tc)
                    res = await self._get_ticker(f"{fc}{tc}")
                    res.from_coin, res.to_coin = from_coin, to_coin
                    log.debug(
                        "Correct Kraken pair for %s/%s is: '%s%s' - caching this pair map into known_pairs",
                        from_coin, to_coin, fc, tc
                    )
                    self.known_pairs[f"{from_coin}_{to_coin}"] = f"{fc}{tc}"
                    return res
                except ExchangeDown:
                    log.debug("Pair %s/%s appears to be invalid. Trying next pair combo.", fc, tc)
        # Our brute force attempt failed. We have no clue what the correct Kraken pair is, whether the exchange
        # is broken, or the pair just outright doesn't exist...
        raise ExchangeDown("Cannot figure out pair for %s/%s or %s is actually down...", from_coin, to_coin, self.name)

    # noinspection PyTypeChecker
    async def _gen_provides(self, *args, **kwargs) -> Set[Tuple[str, str]]:
        t = await self.load_pairs()
        _provides: Set[Tuple[str, str]] = set()
        for k in t:
            _provides.add(tuple(k.split('_')))
        return _provides
    
    # @async_property
    # async def provides(self) -> Set[Tuple[str, str]]:
    #     """
    #     Kraken provides ALL of their tickers through one GET query, so we can generate ``provides``
    #     simply by querying their API via :meth:`.load_pairs` (using :meth:`._gen_provides`)
    #
    #     We cache the provides Set both class-locally in :attr:`._provides`, as well as via the Privex Helpers
    #     Cache system - :mod:`privex.helpers.cache`
    #
    #     """
    #     if empty(self._provides, itr=True):
    #         log.debug("Getting _provides from cache for %s", self.code)
    #         _prov = await self.cache.get(f"pvxex:{self.code}:provides")
    #         if not empty(_prov):
    #             # log.debug("Got _provides from cache")
    #             self._provides = _prov
    #         else:
    #             # log.debug("Calling _gen_provides for %s", self.code)
    #             self._provides = await self._gen_provides()
    #             # log.debug("Adding _provides to cache for %s : %s", self.code, self._provides)
    #             await self.cache.set(f"pvxex:{self.code}:provides", self._provides)
    #     # log.debug("Returning _provides")
    #     return self._provides
    
    async def _get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        return await self.get_ticker(from_coin, to_coin)





from json import JSONDecodeError
from typing import Set, Tuple, Optional, AsyncGenerator, Dict, List, Union

from async_property import async_property
from httpx import HTTPError
from privex.helpers import empty, awaitable_class, r_cache_async, DictObject

from privex.exchange.base import AsyncProvidesAdapter, PriceData
from privex.exchange.exceptions import PairNotFound, ExchangeDown
import httpx
import logging

log = logging.getLogger(__name__)

# cached = adapter_get(AsyncMemoryCache)


# @awaitable_class
class Bittrex(AsyncProvidesAdapter):
    MARKET_API = 'https://api.bittrex.com/v3/markets'
    
    name = "Bittrex"
    code = "bittrex"
    
    _provides = set()
    _extra_provides = set()

    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        prov = await self.provides
        pairs = []
        for f, t in prov:
            pairs += [f"{f.upper()}_{t.upper()}"]

        from_coin, to_coin, = from_coin.upper(), to_coin.upper()
        
        if f"{from_coin}_{to_coin}" in pairs: return True
        if from_coin == 'USD' and f"USDT_{to_coin}" in pairs: return True
        if to_coin == 'USD' and f"{from_coin}_USDT" in pairs: return True

        return False

    @r_cache_async(f"pvxex:bittrex:all_pairs", 300)
    async def load_pairs(self) -> List[str]:
        # First try and get the ticker data from cache
        # ckey = f"pvxex:{self.code}:all_pairs"
        # cdata = await cached.get(ckey)
        #
        # if not empty(cdata):
        #     return cdata
        
        # If we don't have it in the privex-helpers cache, query the Bittrex API, cache the data and return it.
        data = []
        async for from_coin, to_coin in self._load_pairs():
            data += [f"{from_coin}_{to_coin}"]

        # await cached.set(ckey, data)
        
        return data

    async def _load_pairs(self) -> AsyncGenerator[Tuple[str, str], None]:
        """
        Used internally by :meth:`.get_tickers`
        
        Queries the Bittrex market API :attr:`.MARKET_API` asynchronously, and returns
        ``Tuple[str, str]`` objects as an async generator::
        
            >>> async for frm_coin, to_coin in self._load_pairs():
            ...     print(frm_coin, to_coin)
        
        :return AsyncGenerator[PriceData,None] ticker: An async generator of ticker pairs
        """
        # Query the Bittrex API and obtain the pairs data as a list of dictionaries
        async with httpx.AsyncClient() as client:
            data = await client.get(self.MARKET_API, timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
            res: List[dict] = data.json()
        # Loop over each ticker dictionary and return a PriceData object
        for d in res:   # type: dict
            yield d['baseCurrencySymbol'], d['quoteCurrencySymbol']

    # noinspection PyTypeChecker
    async def _gen_provides(self, *args, **kwargs) -> Set[Tuple[str, str]]:
        t = await self.load_pairs()
        _provides: Set[Tuple[str, str]] = set()
        for k in t:
            _provides.add(tuple(k.split('_')))
        return _provides

    # @async_property
    # async def provides(self) -> Set[Tuple[str, str]]:
    #     """
    #     Binance provides ALL of their tickers through one GET query, so we can generate ``provides``
    #     simply by querying their API via :meth:`._load_pairs` (using :meth:`._gen_provides`)
    #
    #     We cache the provides Set both class-locally in :attr:`._provides`, as well as via the Privex Helpers
    #     Cache system - :mod:`privex.helpers.cache`
    #
    #     """
    #     if empty(self._provides, itr=True):
    #         _prov = await self.cache.get(f"pvxex:{self.code}:provides")
    #         if not empty(_prov):
    #             self._provides = _prov
    #         else:
    #             self._provides = await self._gen_provides()
    #             await self.cache.set(f"pvxex:{self.code}:provides", self._provides)
    #     return self._provides

    async def _query(self, from_coin: str, to_coin: str, endpoint='/') -> Union[list, dict]:
        async with httpx.AsyncClient() as client:
            data = await client.get(f"{self.MARKET_API}/{from_coin}-{to_coin}{endpoint}", timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
        return data.json()

    @r_cache_async(lambda self, from_coin, to_coin: f"pvxex:bittrex:tick:{from_coin}:{to_coin}", 30)
    async def _q_ticker(self, from_coin: str, to_coin: str) -> DictObject:
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        res = await self._query(from_coin, to_coin, endpoint='/ticker')
        return DictObject(last=res['lastTradeRate'], bid=res['bidRate'], ask=res['askRate'])

    @r_cache_async(lambda self, from_coin, to_coin: f"pvxex:bittrex:sum:{from_coin}:{to_coin}", 30)
    async def _q_summary(self, from_coin: str, to_coin: str) -> DictObject:
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        res = await self._query(from_coin, to_coin, endpoint='/summary')
        return DictObject(high=res['high'], low=res['low'], volume=res['volume'])

    async def _get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        pairs: List[str] = await self.load_pairs()
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        orig_from, orig_to = from_coin, to_coin
        
        key = f"{from_coin}_{to_coin}"
        if key not in pairs:
            if from_coin == 'USD':
                from_coin, key = "USDT", f"USDT_{to_coin}"
            if to_coin == 'USD':
                to_coin, key = "USDT", f"{from_coin}_USDT"
            
            if key not in pairs:
                raise PairNotFound(f"The coin pair '{from_coin}/{to_coin}' is not supported by {self.name}")
        
        summary = await self._q_summary(from_coin, to_coin)
        ticker = await self._q_ticker(from_coin, to_coin)
        
        return PriceData(
            from_coin=orig_from, to_coin=orig_to, last=ticker.last, bid=ticker.bid, ask=ticker.ask,
            high=summary.high, low=summary.low, volume=summary.volume
        )
        




from privex.helpers import PrivexException


class ExchangeException(PrivexException):
    """Base exception for all exceptions that are part of :mod:`privex.exchange`"""
    pass


class PairNotFound(ExchangeException):
    """Raised when a requested coin pair isn't supported by this exchange adapter"""
    pass


class ProxyMissingPair(PairNotFound):
    """Raised when a requested coin pair required for try_proxy does not exist."""
    pass


class ExchangeDown(ExchangeException):
    """Raised when an exchange appears to be down, e.g. timing out or throwing 4xx / 5xx errors."""
    pass


class ExchangeRateLimited(ExchangeDown):
    """Raised when an exchange adapter encounters a rate limit while querying an exchange"""
    pass



from decimal import Decimal
from typing import List, Optional, Union, Tuple, Set

from privex.helpers import dec_round, empty
from privex.helpers.types import NumberStr
from privex.steemengine import conv_dec
import logging

log = logging.getLogger(__name__)

__all__ = [
    'OptNumStr', 'avg', 'mean', 'med_avg', 'med_avg_out', 'trim_outliers', 'empty_decimal', 'PairListOrSet'
]

OptNumStr = Optional[NumberStr]


def avg(*vals: NumberStr, dp: OptNumStr = '8', rounding: str = None) -> Decimal:
    vals = [conv_dec(v) for v in list(vals) if not empty(v)]
    if len(vals) == 1:
        return vals[0] if empty(dp) else dec_round(vals[0], dp, rounding)
    _avg = conv_dec(sum(vals)) / Decimal(len(vals))
    return _avg if empty(dp) else dec_round(_avg, dp, rounding)


mean = avg


def med_avg_out(*vals: NumberStr, dp: OptNumStr = '8', out_pct: OptNumStr = '50', rounding: str = None) -> Decimal:
    """
    A more advanced version of the standard **median average** mathematical concept.
    
    By default, obtains a standard median average, and then uses ``out_pct`` to filter outlier values which are more than ``out_pct``
    percent bigger or smaller than the median average.
    
    When ``out_pct`` is a valid non-zero number (str, :class:`.Decimal`, float, int):
        
        This function calls :func:`.trim_outliers` to remove any outlier numbers that are more than ``out_pct`` (default: 50%)
        greater or smaller than the median average of ``vals``.
        
        For example, given the numbers ``1, 2, 4, 5, 8, 9``, the median average (midpoint) is 4.5 - and with a 50 ``out_pct``,
        any value smaller than 2.25 (4.5 * 0.5) or larger than 6.75 (4.5 * 1.5) would be removed. The remaining numbers
        are then added together, and divided by the amount of remaining numbers to obtain an average.
        
        However, you can adjust ``out_pct`` to a different percentage, or set it to either ``None`` / ``0`` to disable
        percentage based outlier filtering.
    
    When ``out_pct`` is ``None`` or ``0``:
        
        If there are 3 or less values, then ``vals`` are passed to :func:`.med_avg` to return a simple median average.
        
        If there are 4 to 8 values, then the smallest and largest value are removed.
        
        If there are more than 8 values, then 1/4 of the smallest values and 1/4 of the largest values will be removed.
        An integer 1/4 is used, so for example, 1/4 of 11 values would mean the 2 (2.75 rounded down) largest and smallest
        values are removed from the list.
        
        Once the 1/4 biggest + smallest outliers are removed, the remaining numbers are summed and divided by the number
        of remaining values to obtain an average.
    
    Basic usage - Averaging a mix of decimals and integers (as strings)::
    
        >>> med_avg_out('0.1', '0.2', '2', '2.3', '2.5', '3', '2.1', '8', '15')
        Decimal('2.38000000')
    
    Averaging 9 float values with the default 50% ``out_pct``, and without ``out_pct`` outlier trimming::
    
        >>> med_avg_out(10.10, 60.818, 50.9, 40.111, 45.831, 55.398, 52.155, 90.324, 429.829)
        Decimal('50.86883333')
        >>> med_avg_out(10.10, 60.818, 50.9, 40.111, 45.831, 55.398, 52.155, 90.324, 429.829, out_pct=None)
        Decimal('53.02040000')
    
    Using ``dp`` you can adjust how many decimal places to round the output by (or set it to ``None`` to disable DP rounding)::
    
        >>> med_avg_out(1.211, 3, 5.325, 8.12, dp=2)
        Decimal('4.16')
        >>> med_avg_out(1.211, 3, 5.325, 8.12)
        Decimal('4.16250000')
    
    
    :param NumberStr vals:    Two or more numbers passed as positional arguments to obtain an average for
    
    :param NumberStr dp:      Round (quantize) the resulting :class:`.Decimal` to this many decimal places
    
    :param NumberStr out_pct: When ``out_pct`` isn't empty, outliers will be filtered from ``vals`` if they're more than ``out_pct``
                              percent larger or smaller than the midpoint value. e.g. ``10`` means 10% - if the midpoint was ``50``,
                              then values larger than ``55`` or smaller than ``45`` would be removed.
    
    :param str rounding:      Optional rounding method, e.g. ``ROUND_HALF_DOWN`` or ``ROUND_UP``
    
    :return Decimal avg: The median average of ``vals`` with outliers removed.
    """
    dp = int(dp)
    rate_vals = sorted(list([conv_dec(v) for v in vals]))
    log.debug("got %d rate_vals (before removing outliers): %s", len(rate_vals), rate_vals)
    if not empty(out_pct, zero=True):
        out_pct = conv_dec(out_pct)
        log.debug("out_pct is not empty. trimming outliers using trim_outliers with %f%% tolerance", out_pct)
        rm_out = sorted(trim_outliers(*rate_vals, pct=out_pct))
    # If there's 3 or less values, then we simply return the normal median avg value
    elif len(rate_vals) <= 3:
        log.debug("only %d values (<= 3). returning median average.", len(rate_vals))
        return med_avg(*rate_vals, dp=dp, rounding=rounding)
    
    # Otherwise, we trim away the lowest and highest value(s), depending on how many values there are.
    elif len(rate_vals) <= 8:  # If there are 8 or less values, then we trim away just the first and last value
        log.debug("%d values (<= 8). trimming smallest and largest value.", len(rate_vals))
        rm_out = rate_vals[1:-1]
    else:  # If there are more than 8, we trim away 1/4 of the smallest and largest values.
        qt = int(len(rate_vals) // 4)
        log.debug("%d values (> 8). trimming %d values (1/4) from start and end.", len(rate_vals), qt)
        rm_out = rate_vals[qt:-qt]
    
    log.debug("averaging %d outlier-trimmed values: %s", len(rm_out), rm_out)
    # x = conv_dec(sum(rm_out))
    # mavg = Decimal(x / Decimal(len(rm_out)))
    # mavg = avg(*rm_out, dp=dp, rounding=rounding)
    return avg(*rm_out, dp=dp, rounding=rounding)


def trim_outliers(*vals: NumberStr, pct: NumberStr = '10') -> List[Decimal]:
    """
    Remove outlier values from a set of numbers, based on whether a number is more than ``pct`` percent larger or smaller
    than the median average (midpoint) of the values.
    
    Trimming outliers for integers with the default 10% outlier limit (midpoint is 8700)::
    
        >>> trim_outliers(5000, 7000, 6000, 9000, 9500, 8500, 9700, 8900, 14000, 18000, 2000)
        [Decimal('8500'), Decimal('8900'), Decimal('9000'), Decimal('9500')]
    
    Trimming outliers for decimals with a custom 60% outlier limit (midpoint is 0.425)::
    
        >>> trim_outliers('0.01', '0.4', '0.2', '0.9', '0.35', '4', '0.45', '88.253', '0.6', pct='60')
        [Decimal('0.2'), Decimal('0.35'), Decimal('0.4'), Decimal('0.45'), Decimal('0.6')]
        
    
    :param NumberStr vals:  Multiple numbers specified as positional arguments - to trim outliers from.
    
    :param NumberStr pct:   (kwarg - default: ``10``) A tolerance percentage for outliers, e.g. ``10`` removes any numbers
                            which are 10% larger or smaller than the median average of the numbers.
    
    :return List[Decimal] vals: The original ``vals`` as a list of :class:`.Decimal` 's, minus any outliers
    """
    if len(vals) == 0: raise ValueError("No values passed to trim_outliers... you must pass at least one number!")
    if len(vals) == 1: return [conv_dec(vals[0])]
    pct = conv_dec(pct) / Decimal('100')
    pct_low, pct_high = Decimal('1') - pct, Decimal('1') + pct
    vals = [conv_dec(v) for v in vals]
    mid_val = med_avg(*vals)
    log.debug('trim_outliers midpoint is: %f', mid_val)
    return list(sorted([v for v in vals if (mid_val * pct_low) <= v <= (mid_val * pct_high)]))


def med_avg(*vals: NumberStr, dp: OptNumStr = '8', rounding: str = None) -> Decimal:
    """
    Standard median average.
    
    If there are 3 or less values, the midpoint value will be returned.
    
    If there are 4 or more values, the midpoint, and value before the midpoint will be added together, and then
    divided by two to get the median average.
    
    :param NumberStr vals:   Two or more values to median average.
    :param NumberStr dp:     Decimal places to round to (can be ``None`` to disable rounding)
    :param str rounding:     Optional rounding method, e.g. ``ROUND_HALF_DOWN`` or ``ROUND_UP``
    
    :return Decimal med_avg: The median average of ``vals``
    """
    if len(vals) == 0:
        raise ValueError("No values passed to med_avg... you must pass at least one number!")
    dp = int(dp)
    rate_vals = sorted(list([conv_dec(v) for v in vals]))
    midpoint = int(len(rate_vals) // 2)
    if len(rate_vals) == 1: return rate_vals[0]
    if len(rate_vals) <= 3:
        return rate_vals[midpoint] if empty(dp) else dec_round(rate_vals[midpoint], dp, rounding)
    # mavg = avg(rate_vals[midpoint - 1], rate_vals[midpoint], dp=dp, rounding=rounding)
    # mavg = Decimal((rate_vals[midpoint - 1] + rate_vals[midpoint]) / Decimal('2'))
    # return mavg if empty(dp) else dec_round(mavg, dp, rounding)
    return avg(rate_vals[midpoint - 1], rate_vals[midpoint], dp=dp, rounding=rounding)


def empty_decimal(obj: Optional[Decimal]):
    if empty(obj): return None
    return Decimal(obj)


PairListOrSet = Union[List[Tuple[str, str]], Set[List[Tuple[str, str]]]]



from decimal import Decimal
from json import JSONDecodeError
from typing import Set, Tuple, Optional, AsyncGenerator, Dict, List, Union

from async_property import async_property
from httpx import HTTPError
from privex.helpers import empty, awaitable_class, r_cache_async, DictObject
from privex.exchange.base import AsyncProvidesAdapter, PriceData
from privex.exchange.exceptions import PairNotFound, ExchangeDown
import httpx
import logging

log = logging.getLogger(__name__)
# cached = adapter_get(AsyncMemoryCache)


# @awaitable_class
class CoinGecko(AsyncProvidesAdapter):
    BASE_API = 'https://api.coingecko.com/api/v3'
    
    name = "CoinGecko"
    code = "coingecko"
    
    _provides = set()
    _extra_provides = set()
    
    compare_symbols = ['BTC', 'ETH', 'USD', 'GBP', 'EUR', 'SEK']
    
    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        prov = await self.provides
        pairs = []
        for f, t in prov:
            pairs += [f"{f.upper()}_{t.upper()}"]
        
        from_coin, to_coin, = from_coin.upper(), to_coin.upper()
        
        if f"{from_coin}_{to_coin}" in pairs: return True
        if from_coin == 'USD' and f"USDT_{to_coin}" in pairs: return True
        if to_coin == 'USD' and f"{from_coin}_USDT" in pairs: return True
        
        return False
    
    @r_cache_async(f"pvxex:coingecko:allcoins", 300)
    async def load_coins(self) -> dict:
        # If we don't have it in the privex-helpers cache, query the Coingecko API, cache the data and return it.
        data = {}
        async for c_id, c_name, c_symbol in self._load_coins():
            data[c_symbol.upper()] = dict(id=c_id, name=c_name, c_symbol=c_symbol)
        
        return data
    
    async def _load_coins(self) -> AsyncGenerator[Tuple[str, str], None]:
        """
        Used internally by :meth:`.get_tickers`

        Queries the Bittrex market API :attr:`.MARKET_API` asynchronously, and returns
        ``Tuple[str, str]`` objects as an async generator::

            >>> async for frm_coin, to_coin in self._load_coins():
            ...     print(frm_coin, to_coin)

        :return AsyncGenerator[PriceData,None] ticker: An async generator of ticker pairs
        """
        res: List[dict] = await self._query('coins/list')
        for d in res:  # type: dict
            yield d['id'], d['name'], d['symbol']
    
    # noinspection PyTypeChecker
    async def _gen_provides(self, *args, **kwargs) -> Set[Tuple[str, str]]:
        t = await self.load_coins()
        _provides: Set[Tuple[str, str]] = set()
        for k, v in t.items():
            for sym in self.compare_symbols:
                _provides.add((k.upper(), sym.upper(),))
        return _provides
    
    # @async_property
    # async def provides(self) -> Set[Tuple[str, str]]:
    #     """
    #     Coingecko provides ALL of their tickers through one GET query, so we can generate ``provides``
    #     simply by querying their API via :meth:`._load_pairs` (using :meth:`._gen_provides`)
    #
    #     We cache the provides Set both class-locally in :attr:`._provides`, as well as via the Privex Helpers
    #     Cache system - :mod:`privex.helpers.cache`
    #
    #     """
    #     if empty(self._provides, itr=True):
    #         _prov = await self.cache.get(f"pvxex:{self.code}:provides")
    #         if not empty(_prov):
    #             self._provides = _prov
    #         else:
    #             self._provides = await self._gen_provides()
    #             await self.cache.set(f"pvxex:{self.code}:provides", self._provides)
    #     return self._provides
    
    async def _query(self, endpoint='') -> Union[list, dict]:
        async with httpx.AsyncClient() as client:
            data = await client.get(f"{self.BASE_API}/{endpoint}", timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
        return data.json()

    @r_cache_async(lambda self, from_coin, to_coin: f"pvxex:coingecko:price:{from_coin}:{to_coin}", 30)
    async def _q_price(self, from_coin: str, to_coin: str) -> Decimal:
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        coins = await self.load_coins()
        c = coins[from_coin]
        
        res: dict = await self._query(f"simple/price?ids={c['id']}&vs_currencies={to_coin.lower()}")
        return Decimal(res[c['id']][to_coin.lower()])
    
    async def _get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        _prov = await self.provides
        pairs: List[str] = [f"{f.upper()}_{v.upper()}" for f, v in _prov]
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        orig_from, orig_to = from_coin, to_coin
        
        key = f"{from_coin}_{to_coin}"
        if key not in pairs:
            raise PairNotFound(f"The coin pair '{from_coin}/{to_coin}' is not supported by {self.name}")
        
        return PriceData(
            from_coin=orig_from, to_coin=orig_to, last=await self._q_price(from_coin, to_coin)
        )





from json import JSONDecodeError
from typing import Set, Tuple, Optional, AsyncGenerator, Dict, List
from async_property import async_property
from httpx import HTTPError
from privex.helpers import filter_form
from privex.exchange.base import AsyncProvidesAdapter, PriceData
from privex.exchange.exceptions import PairNotFound, ExchangeDown
import httpx

import logging

log = logging.getLogger(__name__)


class Huobi(AsyncProvidesAdapter):
    TICKER_API = 'https://api.huobi.pro/market/detail/merged?symbol='
    PROVIDES_API = 'https://api.huobi.pro/v1/common/symbols'
    
    name = "Huobi"
    code = "huobi"
    
    _provides = set()
    _extra_provides = set()
    
    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        prov = await self.provides
        pairs = []
        for f, t in prov:
            pairs += [f"{f.upper()}_{t.upper()}"]
        
        from_coin, to_coin, = from_coin.upper(), to_coin.upper()
        
        if f"{from_coin}_{to_coin}" in pairs:
            return True
        
        if from_coin == 'USD' and (f"USDT_{to_coin}" in pairs or f"USDC_{to_coin}" in pairs):
            return True
        if to_coin == 'USD' and (f"{from_coin}_USDT" in pairs or f"{from_coin}_USDC" in pairs):
            return True
        
        return False
    
    # noinspection PyTypeChecker
    async def _gen_provides(self, *args, **kwargs) -> Set[Tuple[str, str]]:
        async with httpx.AsyncClient() as client:
            data = await client.get(self.PROVIDES_API, timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
            res: dict = data.json()
        
        if 'data' not in res:
            raise ExchangeDown(f"'data' missing from {self.name} ({self.code}) PROVIDES API...")
        
        pairs = set()
        for p in res['data']:
            pairs.add((p['base-currency'].upper(), p['quote-currency'].upper()))
        
        return pairs
    
    async def load_pairs(self):
        prov = await self.provides
        res = [f"{f.upper()}_{t.upper()}" for f, t in prov]
        return res

    async def _get_pair(self, from_coin: str, to_coin: str) -> PriceData:
        pairs: List[str] = await self.load_pairs()
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        orig_from, orig_to = from_coin, to_coin
    
        key = f"{from_coin}_{to_coin}"
        if key not in pairs:
            if from_coin == 'USD':
                from_coin, key = "USDT", f"USDT_{to_coin}"
            if to_coin == 'USD':
                to_coin, key = "USDT", f"{from_coin}_USDT"
        
            if key not in pairs:
                raise PairNotFound(f"The coin pair '{from_coin}/{to_coin}' is not supported by {self.name}")
        
        async with httpx.AsyncClient() as client:
            data = await client.get(f"{self.TICKER_API}{from_coin.lower()}{to_coin.lower()}", timeout=20)
            try:
                data.raise_for_status()
            except HTTPError as e:
                try:
                    data = e.response.json()
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {data}")
                except (JSONDecodeError, AttributeError, KeyError):
                    raise ExchangeDown(f"{self.name} appears to be down. Error was: {type(e)} {str(e)}")
            res: dict = data.json()

        if 'tick' not in res:
            raise ExchangeDown(f"'tick' missing from {self.name} ({self.code}) TICKER API...")
        
        # Convert all of the numbers into strings - Huobi returns floats which results in weird
        # rounding issues when converted into Decimal.
        ticker: dict = filter_form(res['tick'], 'close', 'high', 'low', 'amount', 'close', 'open', cast=str)
        ticker['bid'] = str(res['tick']['bid'][0])
        ticker['ask'] = str(res['tick']['ask'][0])
    
        return PriceData(
            from_coin=orig_from, to_coin=orig_to, last=ticker['close'], bid=ticker['bid'], ask=ticker['ask'],
            high=ticker['high'], low=ticker['low'], volume=ticker['amount'], close=ticker['close'], open=ticker['open']
        )




from typing import Set, Tuple, Optional, List
from async_property import async_property
from privex.helpers import empty, r_cache_async, empty_if
from privex.steemengine.exceptions import NoResults
from privex.exchange.base import AsyncProvidesAdapter, PriceData
from privex.exchange.exceptions import PairNotFound
from privex.steemengine import SteemEngineToken, Token
import logging

log = logging.getLogger(__name__)


class SteemEngine(AsyncProvidesAdapter):
    seng = SteemEngineToken(network='steem')
    
    name = "Steem Engine DEX"
    code = "steemengine"
    base_token = 'STEEMP'
    peg_prefix = ''
    peg_suffix = 'P'
    token_limit = 200
    
    _provides = set()
    _extra_provides = set()
    
    linked_symbols = {
        'BTC': 'BTCP',
        'EOS': 'EOSP',
        'TLOS': 'TLOSP',
        'LTC': 'LTCP',
        'DOGE': 'DOGEP',
        'BTS': 'BTSP',
        'SWIFT': 'SWIFTP',
        'STEEM': 'STEEMP',
        'HIVE': 'HIVEP',
    }
    """
    To prevent imposter coins, ``linked_symbols`` contains a map of important "standard" coin symbols, mapped to the
    official pegged version for SteemEngine, ensuring that ``BTC`` for example, never retrieves data for the token ``BTC``,
    since ``BTCP`` is the **real** BTC on SteemEngine
    """
    
    def __init__(self, **ex_settings):
        super().__init__(**ex_settings)
        if 'privex.exchange' not in SteemEngineToken.CACHE_BLACKLIST_MODS:
            SteemEngineToken.CACHE_BLACKLIST_MODS.append('privex.exchange')
        if 'privex.exchange.SteemEngine' not in SteemEngineToken.CACHE_BLACKLIST:
            SteemEngineToken.CACHE_BLACKLIST.append('privex.exchange.SteemEngine')

    @classmethod
    def remove_prefix(cls, token: str) -> str:
        if empty(cls.peg_prefix): return token
        return token[len(cls.peg_prefix):] if token.startswith(cls.peg_prefix) else token

    @classmethod
    def remove_suffix(cls, token: str) -> str:
        if empty(cls.peg_suffix): return token
        return token[:-len(cls.peg_suffix)] if token.endswith(cls.peg_suffix) else token
    
    @classmethod
    def remove_affixes(cls, token: str) -> str:
        return cls.remove_suffix(cls.remove_prefix(token))

    @classmethod
    def _affix_combos(cls, c: str) -> List[str]:
        return [c, cls.remove_prefix(c), cls.remove_suffix(c), cls.remove_affixes(c)]
    
    @property
    def base_token_affixes(self): return self._affix_combos(self.base_token)
    
    @async_property
    async def pairs(self):
        return [f"{f.upper()}_{t.upper()}" for f, t in await self.provides]
    
    @classmethod
    def _find_pair_affix(cls, from_coin: str, to_coin: str, pairs: List[str]) -> Optional[str]:
        from_coin, to_coin, = from_coin.upper(), to_coin.upper()
        from_combos, to_combos = cls._affix_combos(from_coin), cls._affix_combos(to_coin)
        from_combos += [f"{from_coin}{cls.peg_suffix}", f"{cls.peg_prefix}{from_coin}", f"{cls.peg_prefix}{from_coin}{cls.peg_suffix}"]
        for fc in from_combos:
            for tc in to_combos:
                if f"{fc}_{tc}" not in pairs:
                    continue
                return f"{fc}_{tc}"
        return None
    
    async def find_pair_affix(self, from_coin: str, to_coin: str) -> Optional[str]:
        return self._find_pair_affix(from_coin=from_coin, to_coin=to_coin, pairs=await self.pairs)
    
    async def has_pair(self, from_coin: str, to_coin: str) -> bool:
        if to_coin not in self.base_token_affixes:
            return False
        to_coin = self.base_token
        if not empty(await self.find_pair_affix(from_coin=from_coin, to_coin=to_coin)):
            return True
        return False
    
    @async_property
    @r_cache_async(lambda self: f'pvxex:{self.code}:tokens', 600)
    async def tokens(self) -> List[Token]:
        tokens = []
        tokens_loaded = 0
        lim = self.token_limit
        last_res = range(lim)
        batch = 0
        
        while len(last_res) >= lim:
            log.debug("[Batch %d || %d tokens loaded] Loading tokens for %s", batch + 1, tokens_loaded, self.name)
            last_res = self.seng.list_tokens(limit=lim, offset=lim * batch)
            tokens_loaded += len(last_res)
            batch += 1
            tokens += last_res
        
        log.debug("Finished loading %d tokens for %s after %d batches", tokens_loaded, self.name, batch + 1)
        return tokens
    
    @async_property
    @r_cache_async(lambda self: f'pvxex:{self.code}:token_symbols', 600)
    async def token_symbols(self) -> List[str]:
        return list(set([t.symbol for t in await self.tokens]))
    
    async def _gen_provides(self, *args, **kwargs) -> Set[Tuple[str, str]]:
        token_map = set()
        for t in await self.token_symbols:
            t, bt = t.upper(), self.base_token.upper()
            if t == bt:
                continue
            for trim_t in self._affix_combos(t):
                for trim_bt in self._affix_combos(bt):
                    token_map.add((trim_t, trim_bt,))
        return token_map
    
    async def _get_pair(self, from_coin: str, to_coin: str = None) -> PriceData:
        to_coin = empty_if(to_coin, self.base_token)
        from_coin, to_coin = from_coin.upper(), to_coin.upper()
        
        # Avoid imposter tokens by correcting "standard" symbols into their known official symbol on the exchange.
        if from_coin in self.linked_symbols:
            from_coin = self.linked_symbols[from_coin]
        
        if to_coin not in self.base_token_affixes:
            raise PairNotFound(f"Symbol {to_coin} is not supported by {self.name} ({self.code}). Only {self.base_token} is, plus "
                               f"it's symbol variations: {self.base_token_affixes}")
        to_coin = self.base_token
        
        pair = self._find_pair_affix(from_coin, to_coin, await self.pairs)
        if empty(pair):
            raise PairNotFound(f"[No valid pair] Symbol {from_coin} is not supported by {self.name} ({self.code})")
        from_coin = pair.split('_')[0]
        if empty(from_coin):
            raise PairNotFound(f"[from_coin is empty?] Symbol {from_coin} is not supported by {self.name} ({self.code})")
        try:
            ticker = self.seng.get_ticker(from_coin)
        except NoResults:
            raise PairNotFound(f"No ticker found for {from_coin}/{to_coin}")
    
        return PriceData(
            from_coin=from_coin, to_coin=self.base_token, last=ticker.lastPrice, bid=ticker.highestBid,
            ask=ticker.lowestAsk, open=ticker.lastDayPrice, close=ticker.lastDayPrice, volume=ticker.volume
        )

    async def get_pair(self, from_coin: str, to_coin: str = None) -> PriceData:
        to_coin = empty_if(to_coin, self.base_token)
        from_coin = from_coin.upper()
        # Avoid imposter tokens by correcting "standard" symbols into their known official symbol on the exchange.
        if from_coin in self.linked_symbols:
            from_coin = self.linked_symbols[from_coin]
        return await super().get_pair(from_coin=from_coin, to_coin=to_coin)


class HiveEngine(SteemEngine):
    seng = SteemEngineToken(network='hive')
    
    name = "Hive Engine DEX"
    code = "hiveengine"
    base_token = 'SWAP.HIVE'
    peg_prefix = 'SWAP.'
    peg_suffix = ''
    
    _provides = set()
    _extra_provides = set()
    
    linked_symbols = {
        'BTC':   'SWAP.BTC',
        'EOS':   'SWAP.EOS',
        'TLOS':  'SWAP.TLOS',
        'LTC':   'SWAP.LTC',
        'DOGE':  'SWAP.DOGE',
        'BTS':   'SWAP.BTS',
        'SWIFT': 'SWAP.SWIFT',
        'STEEM': 'SWAP.STEEM',
        'HIVE':  'SWAP.HIVE',
    }

